module TestRunner
  module Assert
    class Assertion
      attr_reader :fails
      attr_reader :passes
      attr_accessor :source

      def initialize subject_proc, checks
        @subject_proc = subject_proc
        @checks = checks
        @passes = []
        @fails  = []
        @trace = caller_locations
      end

      def call
        perform_checks
        freeze
        raise to_error if failed?
        subject
      end

      def build_checks
        @checks.map do |check_name, argument|
          TestRunner::Config.internal_logger.debug "Resolving check #{check_name.inspect}, arg=#{argument.inspect}"
          resolve_check check_name, argument
        end
      end

      def failed?
        fails.any?
      end

      def file
        @trace[0].path
      end

      def freeze
        passes.freeze
        fails.freeze
      end

      def line
        @trace[0].lineno
      end

      def passed?
        not failed?
      end

      def perform_checks
        build_checks.each do |check|
          check.evaluate
          ary = if check.passed? then passes else fails end
          ary.push check
        end
      end

      def resolve_check check_name, argument
        check = Checks.resolve check_name do
          check_name = :include if check_name == :includes
          argument = [check_name, *argument]
          Checks[:predicate]
        end
        check.new subject_thunk, argument
      end

      def subject
        subject_thunk.call
      end

      def subject_thunk
        @subject_thunk ||= SubjectThunk.new @subject_proc
      end

      def to_error
        AssertionFailed.new self
      end

      def trace
        BacktraceFilter.(@trace)
      end

      class Refutation < Assertion
        def build_checks
          super.each &:negate
        end
      end

      class SubjectThunk
        def initialize block
          @block = block
        end

        def call
          return @subject if subject_resolved?
          @subject = @block.call
        end

        def expect_error
          raise "called after initially fetched" if subject_resolved?
          @block.call
          nothing_raised = true
        rescue => error
          @subject = error
        ensure
          raise NothingRaised.new if nothing_raised
        end

        def subject_resolved?
          instance_variable_defined? :@subject
        end
      end
    end
  end
end
